

from . import tests00_root
from . import tests01_geom
from . import tests02_data
from . import tests03_openadas
from . import tests04_spectro
from . import tests05_nist
from . import tests06_mesh
from . import tests07_inversions
from . import tests08_diagnostics
from . import tests09_tutorials
